@@ -29,6 +29,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <errno.h>
 
 #include <linux/input.h>
 #include <cairo.h>
@@ -59,6 +60,8 @@ struct virtual_keyboard {
 	uint32_t surrounding_cursor;
 	struct keyboard *keyboard;
 	bool toplevel;
+	bool overlay;
+	struct zwp_input_panel_surface_v1 *ips;
 };
 
 enum key_type {
@@ -374,7 +377,7 @@ redraw_handler(struct widget *widget, void *data)
 	cairo_rectangle(cr, allocation.x, allocation.y, allocation.width, allocation.height);
 	cairo_clip(cr);
 
-	cairo_select_font_face(cr, "sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
+	cairo_select_font_face(cr, "sans-serif", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
 	cairo_set_font_size(cr, 16);
 
 	cairo_translate(cr, allocation.x, allocation.y);
@@ -968,14 +971,38 @@ set_toplevel(struct output *output, struct virtual_keyboard *virtual_keyboard)
 						ZWP_INPUT_PANEL_SURFACE_V1_POSITION_CENTER_BOTTOM);
 
 	virtual_keyboard->toplevel = true;
+	virtual_keyboard->overlay = false;
+	virtual_keyboard->ips = ips;
+}
+
+static void
+set_overlay(struct output *output, struct virtual_keyboard *virtual_keyboard)
+{
+	struct zwp_input_panel_surface_v1 *ips;
+	struct keyboard *keyboard = virtual_keyboard->keyboard;
+
+	ips = zwp_input_panel_v1_get_input_panel_surface(virtual_keyboard->input_panel,
+							 window_get_wl_surface(keyboard->window));
+
+	zwp_input_panel_surface_v1_set_overlay_panel(ips);
+
+	virtual_keyboard->toplevel = false;
+	virtual_keyboard->overlay = true;
+	virtual_keyboard->ips = ips;
 }
 
 static void
 display_output_handler(struct output *output, void *data) {
 	struct virtual_keyboard *keyboard = data;
+	const char *type = getenv("WESTON_KEYBOARD_SURFACE_TYPE");
 
-	if (!keyboard->toplevel)
-		set_toplevel(output, keyboard);
+	if (type && strcasecmp("overlay", type) == 0) {
+		if (!keyboard->overlay)
+			set_overlay(output, keyboard);
+	} else {
+		if (!keyboard->toplevel)
+			set_toplevel(output, keyboard);
+	}
 }
 
 static void
@@ -994,6 +1021,8 @@ keyboard_create(struct virtual_keyboard *virtual_keyboard)
 	virtual_keyboard->keyboard = keyboard;
 
 	window_set_title(keyboard->window, "Virtual keyboard");
+	window_set_appid(keyboard->window,
+			 "org.freedesktop.weston.virtual-keyboard");
 	window_set_user_data(keyboard->window, keyboard);
 
 	widget_set_redraw_handler(keyboard->widget, redraw_handler);
@@ -1010,6 +1039,23 @@ keyboard_create(struct virtual_keyboard *virtual_keyboard)
 					     display_output_handler);
 }
 
+static void
+keyboard_destroy(struct virtual_keyboard *virtual_keyboard)
+{
+	if (virtual_keyboard->ips)
+		zwp_input_panel_surface_v1_destroy(virtual_keyboard->ips);
+
+	if (virtual_keyboard->input_panel)
+		zwp_input_panel_v1_destroy(virtual_keyboard->input_panel);
+
+	if (virtual_keyboard->input_method)
+		zwp_input_method_v1_destroy(virtual_keyboard->input_method);
+
+	widget_destroy(virtual_keyboard->keyboard->widget);
+	window_destroy(virtual_keyboard->keyboard->window);
+	free(virtual_keyboard->keyboard);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -1019,7 +1065,8 @@ main(int argc, char *argv[])
 
 	virtual_keyboard.display = display_create(&argc, argv);
 	if (virtual_keyboard.display == NULL) {
-		fprintf(stderr, "failed to create display: %m\n");
+		fprintf(stderr, "failed to create display: %s\n",
+			strerror(errno));
 		return -1;
 	}
 
@@ -1035,5 +1082,8 @@ main(int argc, char *argv[])
 
 	display_run(virtual_keyboard.display);
 
+	keyboard_destroy(&virtual_keyboard);
+	display_destroy(virtual_keyboard.display);
+
 	return 0;
 }
