@@ -1,6 +1,8 @@
 /*
  * Copyright © 2010 Intel Corporation
  * Copyright © 2013 Jonas Ådahl
+ * Copyright 2017-2018 Collabora, Ltd.
+ * Copyright 2017-2018 General Electric Company
  *
  * Permission is hereby granted, free of charge, to any person obtaining
  * a copy of this software and associated documentation files (the
@@ -33,13 +35,15 @@
 #include <linux/input.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <mtdev.h>
 #include <assert.h>
 #include <libinput.h>
 
-#include "compositor.h"
+#include <libweston/libweston.h>
+#include "backend.h"
+#include "libweston-internal.h"
 #include "libinput-device.h"
 #include "shared/helpers.h"
+#include "shared/timespec-util.h"
 
 void
 evdev_led_update(struct evdev_device *device, enum weston_led weston_leds)
@@ -56,6 +60,21 @@ evdev_led_update(struct evdev_device *device, enum weston_led weston_leds)
 	libinput_device_led_update(device->device, leds);
 }
 
+static void
+ensure_pointer_capability(struct libinput_device *libinput_device)
+{
+	struct evdev_device *device = libinput_device_get_user_data(libinput_device);
+	struct weston_seat *seat = device->seat;
+
+	if (!libinput_device_has_capability(libinput_device, LIBINPUT_DEVICE_CAP_POINTER))
+		return;
+
+	if (!(device->seat_caps & EVDEV_SEAT_POINTER)) {
+		weston_seat_init_pointer(seat);
+		device->seat_caps |= EVDEV_SEAT_POINTER;
+	}
+}
+
 static void
 handle_keyboard_key(struct libinput_device *libinput_device,
 		    struct libinput_event_keyboard *keyboard_event)
@@ -66,6 +85,7 @@ handle_keyboard_key(struct libinput_device *libinput_device,
 		libinput_event_keyboard_get_key_state(keyboard_event);
 	int seat_key_count =
 		libinput_event_keyboard_get_seat_key_count(keyboard_event);
+	struct timespec time;
 
 	/* Ignore key events that are not seat wide state changes. */
 	if ((key_state == LIBINPUT_KEY_STATE_PRESSED &&
@@ -74,8 +94,10 @@ handle_keyboard_key(struct libinput_device *libinput_device,
 	     seat_key_count != 0))
 		return;
 
-	notify_key(device->seat,
-		   libinput_event_keyboard_get_time(keyboard_event),
+	timespec_from_usec(&time,
+			   libinput_event_keyboard_get_time_usec(keyboard_event));
+
+	notify_key(device->seat, &time,
 		   libinput_event_keyboard_get_key(keyboard_event),
 		   key_state, STATE_UPDATE_AUTOMATIC);
 }
@@ -87,26 +109,27 @@ handle_pointer_motion(struct libinput_device *libinput_device,
 	struct evdev_device *device =
 		libinput_device_get_user_data(libinput_device);
 	struct weston_pointer_motion_event event = { 0 };
-	uint64_t time_usec =
-		libinput_event_pointer_get_time_usec(pointer_event);
+	struct timespec time;
 	double dx_unaccel, dy_unaccel;
 
+	ensure_pointer_capability(libinput_device);
+
+	timespec_from_usec(&time,
+			   libinput_event_pointer_get_time_usec(pointer_event));
 	dx_unaccel = libinput_event_pointer_get_dx_unaccelerated(pointer_event);
 	dy_unaccel = libinput_event_pointer_get_dy_unaccelerated(pointer_event);
 
 	event = (struct weston_pointer_motion_event) {
 		.mask = WESTON_POINTER_MOTION_REL |
 			WESTON_POINTER_MOTION_REL_UNACCEL,
-		.time_usec = time_usec,
+		.time = time,
 		.dx = libinput_event_pointer_get_dx(pointer_event),
 		.dy = libinput_event_pointer_get_dy(pointer_event),
 		.dx_unaccel = dx_unaccel,
 		.dy_unaccel = dy_unaccel,
 	};
 
-	notify_motion(device->seat,
-		      libinput_event_pointer_get_time(pointer_event),
-		      &event);
+	notify_motion(device->seat, &time, &event);
 
 	return true;
 }
@@ -119,14 +142,17 @@ handle_pointer_motion_absolute(
 	struct evdev_device *device =
 		libinput_device_get_user_data(libinput_device);
 	struct weston_output *output = device->output;
-	uint32_t time;
+	struct timespec time;
 	double x, y;
 	uint32_t width, height;
 
+	ensure_pointer_capability(libinput_device);
+
 	if (!output)
 		return false;
 
-	time = libinput_event_pointer_get_time(pointer_event);
+	timespec_from_usec(&time,
+			   libinput_event_pointer_get_time_usec(pointer_event));
 	width = device->output->current_mode->width;
 	height = device->output->current_mode->height;
 
@@ -136,7 +162,7 @@ handle_pointer_motion_absolute(
 							      height);
 
 	weston_output_transform_coordinate(device->output, x, y, &x, &y);
-	notify_motion_absolute(device->seat, time, x, y);
+	notify_motion_absolute(device->seat, &time, x, y);
 
 	return true;
 }
@@ -151,6 +177,9 @@ handle_pointer_button(struct libinput_device *libinput_device,
 		libinput_event_pointer_get_button_state(pointer_event);
 	int seat_button_count =
 		libinput_event_pointer_get_seat_button_count(pointer_event);
+	struct timespec time;
+
+	ensure_pointer_capability(libinput_device);
 
 	/* Ignore button events that are not seat wide state changes. */
 	if ((button_state == LIBINPUT_BUTTON_STATE_PRESSED &&
@@ -159,10 +188,12 @@ handle_pointer_button(struct libinput_device *libinput_device,
 	     seat_button_count != 0))
 		return false;
 
-	notify_button(device->seat,
-		      libinput_event_pointer_get_time(pointer_event),
-		      libinput_event_pointer_get_button(pointer_event),
-                      button_state);
+	timespec_from_usec(&time,
+			   libinput_event_pointer_get_time_usec(pointer_event));
+
+	notify_button(device->seat, &time,
+	              libinput_event_pointer_get_button(pointer_event),
+	              button_state);
 
 	return true;
 }
@@ -227,6 +258,9 @@ handle_pointer_axis(struct libinput_device *libinput_device,
 	enum libinput_pointer_axis_source source;
 	uint32_t wl_axis_source;
 	bool has_vert, has_horiz;
+	struct timespec time;
+
+	ensure_pointer_capability(libinput_device);
 
 	has_vert = libinput_event_pointer_has_axis(pointer_event,
 				   LIBINPUT_POINTER_AXIS_SCROLL_VERTICAL);
@@ -257,6 +291,9 @@ handle_pointer_axis(struct libinput_device *libinput_device,
 
 	notify_axis_source(device->seat, wl_axis_source);
 
+	timespec_from_usec(&time,
+			   libinput_event_pointer_get_time_usec(pointer_event));
+
 	if (has_vert) {
 		axis = LIBINPUT_POINTER_AXIS_SCROLL_VERTICAL;
 		vert_discrete = get_axis_discrete(pointer_event, axis);
@@ -267,9 +304,7 @@ handle_pointer_axis(struct libinput_device *libinput_device,
 		weston_event.discrete = vert_discrete;
 		weston_event.has_discrete = (vert_discrete != 0);
 
-		notify_axis(device->seat,
-			    libinput_event_pointer_get_time(pointer_event),
-			    &weston_event);
+		notify_axis(device->seat, &time, &weston_event);
 	}
 
 	if (has_horiz) {
@@ -282,14 +317,124 @@ handle_pointer_axis(struct libinput_device *libinput_device,
 		weston_event.discrete = horiz_discrete;
 		weston_event.has_discrete = (horiz_discrete != 0);
 
-		notify_axis(device->seat,
-			    libinput_event_pointer_get_time(pointer_event),
-			    &weston_event);
+		notify_axis(device->seat, &time, &weston_event);
 	}
 
 	return true;
 }
 
+static struct weston_output *
+touch_get_output(struct weston_touch_device *device)
+{
+	struct evdev_device *evdev_device = device->backend_data;
+
+	return evdev_device->output;
+}
+
+static const char *
+touch_get_calibration_head_name(struct weston_touch_device *device)
+{
+	struct evdev_device *evdev_device = device->backend_data;
+	struct weston_output *output = evdev_device->output;
+	struct weston_head *head;
+
+	if (!output)
+		return NULL;
+
+	assert(output->enabled);
+	if (evdev_device->output_name)
+		return evdev_device->output_name;
+
+	/* No specific head was configured, so the association was made by
+	 * the default rule. Just grab whatever head's name.
+	 */
+	wl_list_for_each(head, &output->head_list, output_link)
+		return head->name;
+
+	assert(0);
+	return NULL;
+}
+
+static void
+touch_get_calibration(struct weston_touch_device *device,
+		      struct weston_touch_device_matrix *cal)
+{
+	struct evdev_device *evdev_device = device->backend_data;
+
+	libinput_device_config_calibration_get_matrix(evdev_device->device,
+						      cal->m);
+}
+
+static void
+do_set_calibration(struct evdev_device *evdev_device,
+		   const struct weston_touch_device_matrix *cal)
+{
+	enum libinput_config_status status;
+
+	weston_log("input device %s: applying calibration:\n",
+		   libinput_device_get_sysname(evdev_device->device));
+	weston_log_continue(STAMP_SPACE "  %f %f %f\n",
+			    cal->m[0], cal->m[1], cal->m[2]);
+	weston_log_continue(STAMP_SPACE "  %f %f %f\n",
+			    cal->m[3], cal->m[4], cal->m[5]);
+
+	status = libinput_device_config_calibration_set_matrix(evdev_device->device,
+							       cal->m);
+	if (status != LIBINPUT_CONFIG_STATUS_SUCCESS)
+		weston_log("Error: Failed to apply calibration.\n");
+}
+
+static void
+touch_set_calibration(struct weston_touch_device *device,
+		      const struct weston_touch_device_matrix *cal)
+{
+	struct evdev_device *evdev_device = device->backend_data;
+
+	/* Stop output hotplug from reloading the WL_CALIBRATION values.
+	 * libinput will maintain the latest calibration for us.
+	 */
+	evdev_device->override_wl_calibration = true;
+
+	do_set_calibration(evdev_device, cal);
+}
+
+static const struct weston_touch_device_ops touch_calibration_ops = {
+	.get_output = touch_get_output,
+	.get_calibration_head_name = touch_get_calibration_head_name,
+	.get_calibration = touch_get_calibration,
+	.set_calibration = touch_set_calibration
+};
+
+static struct weston_touch_device *
+create_touch_device(struct evdev_device *device)
+{
+	const struct weston_touch_device_ops *ops = NULL;
+	struct weston_touch_device *touch_device;
+	struct udev_device *udev_device;
+
+	if (libinput_device_config_calibration_has_matrix(device->device))
+		ops = &touch_calibration_ops;
+
+	udev_device = libinput_device_get_udev_device(device->device);
+	if (!udev_device)
+		return NULL;
+
+	touch_device = weston_touch_create_touch_device(device->seat->touch_state,
+					udev_device_get_syspath(udev_device),
+					device, ops);
+
+	udev_device_unref(udev_device);
+
+	if (!touch_device)
+		return NULL;
+
+	weston_log("Touchscreen - %s - %s\n",
+		   libinput_device_get_name(device->device),
+		   touch_device->syspath);
+
+	return touch_device;
+}
+
 static void
 handle_touch_with_coords(struct libinput_device *libinput_device,
 			 struct libinput_event_touch *touch_event,
@@ -299,14 +444,16 @@ handle_touch_with_coords(struct libinput_device *libinput_device,
 		libinput_device_get_user_data(libinput_device);
 	double x;
 	double y;
+	struct weston_point2d_device_normalized norm;
 	uint32_t width, height;
-	uint32_t time;
+	struct timespec time;
 	int32_t slot;
 
 	if (!device->output)
 		return;
 
-	time = libinput_event_touch_get_time(touch_event);
+	timespec_from_usec(&time,
+			   libinput_event_touch_get_time_usec(touch_event));
 	slot = libinput_event_touch_get_seat_slot(touch_event);
 
 	width = device->output->current_mode->width;
@@ -317,7 +464,14 @@ handle_touch_with_coords(struct libinput_device *libinput_device,
 	weston_output_transform_coordinate(device->output,
 					   x, y, &x, &y);
 
-	notify_touch(device->seat, time, slot, x, y, touch_type);
+	if (weston_touch_device_can_calibrate(device->touch_device)) {
+		norm.x = libinput_event_touch_get_x_transformed(touch_event, 1);
+		norm.y = libinput_event_touch_get_y_transformed(touch_event, 1);
+		notify_touch_normalized(device->touch_device, &time, slot,
+					x, y, &norm, touch_type);
+	} else {
+		notify_touch(device->touch_device, &time, slot, x, y, touch_type);
+	}
 }
 
 static void
@@ -340,10 +494,13 @@ handle_touch_up(struct libinput_device *libinput_device,
 {
 	struct evdev_device *device =
 		libinput_device_get_user_data(libinput_device);
-	uint32_t time = libinput_event_touch_get_time(touch_event);
+	struct timespec time;
 	int32_t slot = libinput_event_touch_get_seat_slot(touch_event);
 
-	notify_touch(device->seat, time, slot, 0, 0, WL_TOUCH_UP);
+	timespec_from_usec(&time,
+			   libinput_event_touch_get_time_usec(touch_event));
+
+	notify_touch(device->touch_device, &time, slot, 0, 0, WL_TOUCH_UP);
 }
 
 static void
@@ -352,9 +509,8 @@ handle_touch_frame(struct libinput_device *libinput_device,
 {
 	struct evdev_device *device =
 		libinput_device_get_user_data(libinput_device);
-	struct weston_seat *seat = device->seat;
 
-	notify_touch_frame(seat);
+	notify_touch_frame(device->touch_device);
 }
 
 int
@@ -367,6 +523,9 @@ evdev_device_process_event(struct libinput_event *event)
 	int handled = 1;
 	bool need_frame = false;
 
+	if (!device)
+		return 0;
+
 	switch (libinput_event_get_type(event)) {
 	case LIBINPUT_EVENT_KEYBOARD_KEY:
 		handle_keyboard_key(libinput_device,
@@ -424,16 +583,8 @@ notify_output_destroy(struct wl_listener *listener, void *data)
 	struct evdev_device *device =
 		container_of(listener,
 			     struct evdev_device, output_destroy_listener);
-	struct weston_compositor *c = device->seat->compositor;
-	struct weston_output *output;
 
-	if (!device->output_name && !wl_list_empty(&c->output_list)) {
-		output = container_of(c->output_list.next,
-				      struct weston_output, link);
-		evdev_device_set_output(device, output);
-	} else {
-		device->output = NULL;
-	}
+	evdev_device_set_output(device, NULL);
 }
 
 /**
@@ -450,24 +601,37 @@ evdev_device_set_calibration(struct evdev_device *device)
 	const char *sysname = libinput_device_get_sysname(device->device);
 	const char *calibration_values;
 	uint32_t width, height;
-	float calibration[6];
-	enum libinput_config_status status;
+	struct weston_touch_device_matrix calibration;
 
-	if (!device->output)
+	if (!libinput_device_config_calibration_has_matrix(device->device))
 		return;
 
+	/* If LIBINPUT_CALIBRATION_MATRIX was set to non-identity, we will not
+	 * override it with WL_CALIBRATION. It also means we don't need an
+	 * output to load a calibration. */
+	if (libinput_device_config_calibration_get_default_matrix(
+							  device->device,
+							  calibration.m) != 0)
+		return;
+
+	/* touch_set_calibration() has updated the values, do not load old
+	 * values from WL_CALIBRATION.
+	 */
+	if (device->override_wl_calibration)
+		return;
+
+	if (!device->output) {
+		weston_log("input device %s has no enabled output associated "
+			   "(%s named), skipping calibration for now.\n",
+			   sysname, device->output_name ?: "none");
+		return;
+	}
+
 	width = device->output->width;
 	height = device->output->height;
 	if (width == 0 || height == 0)
 		return;
 
-	/* If libinput has a pre-set calibration matrix, don't override it */
-	if (!libinput_device_config_calibration_has_matrix(device->device) ||
-	    libinput_device_config_calibration_get_default_matrix(
-							  device->device,
-							  calibration) != 0)
-		return;
-
 	udev = udev_new();
 	if (!udev)
 		return;
@@ -482,37 +646,35 @@ evdev_device_set_calibration(struct evdev_device *device)
 		udev_device_get_property_value(udev_device,
 					       "WL_CALIBRATION");
 
+	if (calibration_values) {
+		weston_log("Warning: input device %s has WL_CALIBRATION property set. "
+			   "Support for it will be removed in the future. "
+			   "Please use LIBINPUT_CALIBRATION_MATRIX instead.\n",
+			   sysname);
+	}
+
 	if (!calibration_values || sscanf(calibration_values,
 					  "%f %f %f %f %f %f",
-					  &calibration[0],
-					  &calibration[1],
-					  &calibration[2],
-					  &calibration[3],
-					  &calibration[4],
-					  &calibration[5]) != 6)
+					  &calibration.m[0],
+					  &calibration.m[1],
+					  &calibration.m[2],
+					  &calibration.m[3],
+					  &calibration.m[4],
+					  &calibration.m[5]) != 6)
 		goto out;
 
-	weston_log("Applying calibration: %f %f %f %f %f %f "
-		   "(normalized %f %f)\n",
-		    calibration[0],
-		    calibration[1],
-		    calibration[2],
-		    calibration[3],
-		    calibration[4],
-		    calibration[5],
-		    calibration[2] / width,
-		    calibration[5] / height);
-
 	/* normalize to a format libinput can use. There is a chance of
 	   this being wrong if the width/height don't match the device
 	   width/height but I'm not sure how to fix that */
-	calibration[2] /= width;
-	calibration[5] /= height;
+	calibration.m[2] /= width;
+	calibration.m[5] /= height;
 
-	status = libinput_device_config_calibration_set_matrix(device->device,
-							       calibration);
-	if (status != LIBINPUT_CONFIG_STATUS_SUCCESS)
-		weston_log("Failed to apply calibration.\n");
+	do_set_calibration(device, &calibration);
+
+	weston_log_continue(STAMP_SPACE "  raw translation %f %f for output %s\n",
+		   calibration.m[2] * width,
+		   calibration.m[5] * height,
+		   device->output->name);
 
 out:
 	if (udev_device)
@@ -524,11 +686,28 @@ void
 evdev_device_set_output(struct evdev_device *device,
 			struct weston_output *output)
 {
+	if (device->output == output)
+		return;
+
 	if (device->output_destroy_listener.notify) {
 		wl_list_remove(&device->output_destroy_listener.link);
 		device->output_destroy_listener.notify = NULL;
 	}
 
+	if (!output) {
+		weston_log("output for input device %s removed\n",
+			   libinput_device_get_sysname(device->device));
+
+		device->output = NULL;
+		return;
+	}
+
+	weston_log("associating input device %s with output %s "
+		   "(%s by udev)\n",
+		   libinput_device_get_sysname(device->device),
+		   output->name,
+		   device->output_name ?: "none");
+
 	device->output = output;
 	device->output_destroy_listener.notify = notify_output_destroy;
 	wl_signal_add(&output->destroy_signal,
@@ -552,18 +731,29 @@ evdev_device_create(struct libinput_device *libinput_device,
 
 	if (libinput_device_has_capability(libinput_device,
 					   LIBINPUT_DEVICE_CAP_KEYBOARD)) {
-		weston_seat_init_keyboard(seat, NULL);
+		if (weston_seat_init_keyboard(seat, NULL) < 0) {
+			free(device);
+			return NULL;
+		}
+
 		device->seat_caps |= EVDEV_SEAT_KEYBOARD;
 	}
-	if (libinput_device_has_capability(libinput_device,
-					   LIBINPUT_DEVICE_CAP_POINTER)) {
-		weston_seat_init_pointer(seat);
-		device->seat_caps |= EVDEV_SEAT_POINTER;
-	}
+
 	if (libinput_device_has_capability(libinput_device,
 					   LIBINPUT_DEVICE_CAP_TOUCH)) {
-		weston_seat_init_touch(seat);
+		if (weston_seat_init_touch(seat) < 0) {
+			/* maybe we're a keyboard + touch device thus we need
+			 * to release the keyboard in case we couldn't make use
+			 * of the touch */
+			if (device->seat_caps & EVDEV_SEAT_KEYBOARD)
+				weston_seat_release_keyboard(seat);
+
+			free(device);
+			return NULL;
+		}
+
 		device->seat_caps |= EVDEV_SEAT_TOUCH;
+		device->touch_device = create_touch_device(device);
 	}
 
 	libinput_device_set_user_data(libinput_device, device);
@@ -579,14 +769,15 @@ evdev_device_destroy(struct evdev_device *device)
 		weston_seat_release_pointer(device->seat);
 	if (device->seat_caps & EVDEV_SEAT_KEYBOARD)
 		weston_seat_release_keyboard(device->seat);
-	if (device->seat_caps & EVDEV_SEAT_TOUCH)
+	if (device->seat_caps & EVDEV_SEAT_TOUCH) {
+		weston_touch_device_destroy(device->touch_device);
 		weston_seat_release_touch(device->seat);
+	}
 
 	if (device->output)
 		wl_list_remove(&device->output_destroy_listener.link);
 	wl_list_remove(&device->link);
 	libinput_device_unref(device->device);
-	free(device->devnode);
 	free(device->output_name);
 	free(device);
 }
